    /**********************************************************************
    * This source code is public domain, and may be freely distributed,	  *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.						  *
    **********************************************************************/

/*****************************************************************************
*  smi_to_2d.c -- read SMILES, write tdt continaing 2D data item.
*
*  Reads SMILES from standard input, writes tdt's (Thor datatrees) to
*  standard output and errors to standard error, e.g.:
*
*     $ smi_to_2d < my.smi > my.tdt 2> my.err
*
*  To include all hydrogen coordinates, use "-h", e.g.
*
*     $ smi_to_2d -h < my.smi > my.tdt 2> my.err
*
*  This program might be useful for adding 2D coordinates to a database
*  (its output can be loaded into a database, e.g. with thorload or
*  sthorman).
*
*  The source code shows how to create a depiction object and calculate its
*  coordinates.  Since this program does not do graphical output, it is
*  linked to a dummy drawing library (dl_stubs).
*
*  THOR's 2D format treats "implicit" and "explicit" hydrogens differently.
*  "Implicit" hydrogens are the "ordinary" ones that don't appear in a
*  SMILES, i.e. they are singly bonded to a "heavy" (non-hydrogen) atom,
*  have mass 1, have no charge, and are not attached to a chiral center.
*  Any hydrogen that doesn't meet these conditions, i.e. that would appear
*  in the unique SMILES, is an "explicit" hydrogen.
*
*  In the 2D format used by THOR, all "heavy" (non-H) atoms, plus all
*  explicit hydrogens, must have their coordinates specified.  They are
*  listed in unique-SMILES order.  You can optionally specify the implicit
*  hydrogens' coordinates; since implicit hydrogens don't appear in the
*  SMILES, these are appended to the end of the heavy-atom coordinates in
*  the order of the heavy atoms to which they are attached.  Although this
*  sounds tricky at first, it is made easy by virtue of the function
*  dt_canatom_stream(3): its "addh" (third) parameter causes hydrogens to
*  be added to the end of the stream in the exact order that we need them
*  for this function.
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include "dt_smiles.h"
#include "dt_smarts.h"
#include "dt_depict.h"
#include "du_utils.h"

#define MAXSMI 2000

/***************************************************************************
* FUNCTION: usage
*
* DESCRIPTION:
*	Prints "usage" message and exits.
***************************************************************************/

static void usage(char *myname)
{
  fprintf(stderr, "usage: %s [--help][-h][-a tdt]\n", myname);
  fprintf(stderr, "   -h:  print coordinates for all hydrogen atoms\n");
  fprintf(stderr, "        (hydrogen coordinates are normally suppressed)\n");
  fprintf(stderr, "   -a tdt: align coordinates to core smiles\n");
  fprintf(stderr, "          e.g: -a '$SMI<CCC>2D<0.88,1.53,0.00,2.04,-0.88,1.53>|'\n");
  exit(1);
}

/***************************************************************************
* FUNCTION: fillh
*
* DESCRIPTION:
*	Convert implicit hydrogens to explicit in a molecule.  That is,
*	for every "heavy" atom, replace its implicit-hydrogen count with
*	actual hydrogen-atom objects.
*
*	In order to get the coordinates for "implicit" hydroges (those that
*	don't appear in the SMILES), we have to convert them to "explicit"
*	hydrogen atoms.  Once they are explicit, the function
*	dt_canatom_stream(mol, 1, 1) will return a stream that includes all
*	atoms, "heavy" and hydrogen.  This allows us to get the coordinates
*	of the hydrogen atoms, which would otherwise be unavailable.
*
*	The function dt_mod_off() dealloc's the molecule on failure.  So
*	for consistency, we deallocate the molecule if this function
*	fails for any reason.
*
* RETURNS: 0 if it fails, 1 if it works.
***************************************************************************/

static int fillh(dt_Handle mol)
{
  dt_Handle atom, h, atoms, s;
  int hcount;

  /*** Copy atoms to a sequence ***/
  if (   NULL_OB == (s = dt_stream(mol, TYP_ATOM))
      || NULL_OB == (atoms = dt_alloc_seq())) {
    dt_errorsave("fillh", DX_ERR_ERROR, "Out of memory");
    dt_dealloc(mol);
    return 0;
  }
  while (NULL_OB != (atom = dt_next(s)))
    if (NULL_OB == dt_append(atoms, atom)) {
      dt_errorsave("fillh", DX_ERR_ERROR, "Out of memory");
      dt_dealloc(atoms);
      dt_dealloc(mol);
      return 0;
    }
  dt_dealloc(s);

  /*** Convert implicit to explicit ***/
  if (!dt_mod_on(mol)) {
    dt_errorsave("fillh", DX_ERR_ERROR, "dt_mod_on() failed!");
    dt_dealloc(atoms);
    dt_dealloc(mol);
    return 0;
  }
  dt_reset(atoms);
  while (NULL_OB != (atom = dt_next(atoms))) {
    hcount = dt_imp_hcount(atom);
    dt_setimp_hcount(atom, 0);
    while (hcount-- > 0) {
      if (   NULL_OB == (h = dt_addatom(mol, 1, 0))
	  || !dt_addbond(atom, h, DX_BTY_SINGLE)) {
	dt_errorsave("fillh", DX_ERR_ERROR,
		     "dt_addatom() or dt_addbond() failed!");
	dt_dealloc(atoms);
	dt_dealloc(mol);
	return 0;
      }
    }
  }
  dt_dealloc(atoms);
  return dt_mod_off(mol);
}

char *id_matched_atoms (dt_Handle mol, dt_Handle matched) {
/* returns a string with each char position representing
   an atom;  0 means that atom was not matched, 1 means it was */

 int natoms;
 dt_Handle atoms;
 dt_Handle atom;
 dt_Handle paths;
 dt_Handle path;
 int i;
 char *s;

 natoms = dt_count(mol, TYP_ATOM);
 s = (char *) malloc(natoms+1);
 s[0] = '\0';
 atoms  = dt_canatom_stream(mol, 1, 0);
 i = 0;
 paths = dt_stream(matched, TYP_PATH);
 path = dt_next(paths); /* should only be one; used dt_umatch */
 while ( NULL_OB != (atom = dt_next(atoms)) ) {
  if (dt_member(path,atom)) {
   s[i++] = '1';
  } else {
   s[i++] = '0';
  }
 }
 s[i] = '\0';
 dt_dealloc (paths);
 dt_dealloc (atoms);
 return (s);
}

int set_coords (dt_Handle dep, dt_Handle mol, dt_Handle coremol,
  dt_Handle matched, float corex[], float corey[]) {

/* set the coords of the atoms in mol to those in the coremol;
   coremol is a smarts handle;
   matched is the pathset from dt_umatch for coremol in mol;
   corex and corey are atom coords, in orig coremol/smi order.
*/

 dt_Handle atoms;
 dt_Handle atom;
 dt_Handle path;
 dt_Handle paths;
 int i=0;

  paths = dt_stream(matched, TYP_PATH);
  while ( NULL_OB != (path=dt_next(paths)) ) {
   atoms = dt_stream(path, TYP_ATOM);
   i=0;
   while ( NULL_OB != (atom = dt_next(atoms)) ) {
    /* fprintf (stderr, "%d\n", atom); */
    if ( !dt_setcoord(dep, atom, corex[i], corey[i], 0.0) ) { i=-1; break; }
    i++;
   }
  }
  dt_dealloc (atoms);
  dt_dealloc (paths);
  return(i);
}

/***************************************************************************
* FUNCTION: main
***************************************************************************/

main(int argc, char **argv)
{
  dt_Handle mol, dep, atom, atoms, coremol, matched;
  char smi[MAXSMI], *usmi, *ismi, *buf, *b, *p;
  int  lenu, leni, bsize, addh;
  float x, y, z;
  int i;
  int aligned;
  char tdt[500];
  char *core;
  char *core2d;
  int corelen = 0;
  int core2dlen = 0;
  float corex[100];
  float corey[100];
  char *c;
  char *amatch = NULL;
  int nset;

  addh = 0;
  aligned = 0;
  /*** Check for command line args ***/
  for (i=1; i<argc; i++) {

   if ( !strcmp(argv[i], "--help") ) {
    usage(argv[0]);

   } else if ( !strcmp(argv[i], "-h") ) {
    addh = 1;

   } else if ( !strcmp(argv[i], "-a") ) {
    aligned = 1;
/* next arg after -a is a TDT; should have $SMI<>, optional 2D<> */
    strncpy(tdt, argv[++i], sizeof(tdt));
    core = du_find_dataitem(&corelen, strlen(tdt), tdt, 4, "$SMI");
    if (NULL_OB == core) {
     aligned = 0;
     fprintf(stderr, "no alignment: can't find $SMI in %s\n", tdt);
    }
    core2d = du_find_dataitem(&core2dlen, strlen(tdt), tdt, 2, "2D");
/* parse x,y coords out of the 2D<> tag */
    if (NULL_OB != core2d) {
     i=0;
     c = strtok(core2d+3, ",");
     while ( c ) {
      corex[i] = atof(c);
      c = strtok(NULL, ",");
      corey[i] = atof(c);
      c = strtok(NULL, ",");
      i++;
     }
    } else {
/* should auto get 2D coords here */
     aligned = 0;
    }

   } else {
    usage(argv[0]);
   }

  }

/* check validity of TDT arg */
  if ( aligned && NULL_OB == (coremol =  dt_smilin(corelen-5-1,core+5)) ) {
   fprintf(stderr, "ignored invalid $SMI in argument %s\n", tdt);
   aligned = 0;
  } else {
   dt_dealloc(coremol);
   coremol =  dt_smartin(corelen-5-1,core+5);
  }

  /*** Loop over input ***/

  while( fgets(smi,2000,stdin) ) {
    smi[strlen(smi)-1]='\0';
    /*** Parse SMILES, create depiction, compute (x,y) coordinates. ***/

    if (NULL_OB == (mol = dt_smilin(strlen(smi), smi))) {
      fprintf(stderr, "%s: can't process '%s'\n", argv[0], smi);
      du_printsmileserrors(stderr);
      dt_errorclear();
      continue;
    }

    /*** If we're doing H's, convert implicit hydrogens to explicit ***/

    if (addh && !fillh(mol)) {
      fprintf(stderr, "%s: can't process '%s'\n", argv[0], smi);
      du_printerrors(stderr, 2);
      continue;
    }

    /*** Allocate depiction ***/

    if (   NULL_OB == (dep = dt_alloc_depiction(mol))) {
      fprintf(stderr, "%s: Failed to create depiction for '%s'\n",
	      argv[0], smi);
      du_printerrors(stderr, 2);
      dt_dealloc(mol);
      continue;
    }
    /*** set matched atoms' coords to aligned coords, if opted ***/
    if ( aligned ) {
     if ( NULL_OB != (matched = dt_umatch(coremol, mol, 1)) ) {
      if ( (nset=set_coords(dep, mol, coremol, matched, corex, corey)) < 1 ) {
       fprintf(stderr, "%s: Failed to set coords\n", argv[0]);
      } else {
       dt_setboolean(dep, 6, "use_2d", 1);
       amatch = id_matched_atoms(mol, matched);
       /* fprintf(stderr, "%s: set %d coords\n", argv[0], nset); */
      }
     }
    }

    /*** compute 2D coords for this depiction ***/

    if ( !dt_calcxy(dep) ) {
      fprintf(stderr, "%s: Failed to compute 2D coords for '%s'\n",
	      argv[0], smi);
      du_printerrors(stderr, 2);
      dt_dealloc(mol);
      continue;
    }

    /*** Write unique SMILES as tree root. ***/

    usmi = dt_cansmiles(&lenu, mol, 0);
    printf("$SMI<%.*s>", lenu, usmi);

    /*** Add $ISM as 2D's identifier if SMILES is isomeric. ***/

    ismi = dt_cansmiles(&leni, mol, 1);
    if (leni != lenu || 0 != memcmp(ismi, usmi, lenu))
      printf("$ISM<%.*s>", leni, ismi);


    /*** Create 2D data item. Note how "addh" causes H's to be included ***/

    atoms   = dt_canatom_stream(mol, 1, addh);
    bsize   = 16 * dt_count(mol, TYP_ATOM) + 1;
    b = buf = (char *) malloc(bsize);
    for (b = buf; NULL_OB != (atom = dt_next(atoms)); b += 16) {
      dt_getcoord(dep, atom, &x, &y, &z);
      sprintf(b, ",%7.2f,%7.2f", x, y);
    }

    /*** Remove blanks to save space. ***/

    for (p = b = buf; bsize--; b++)
      if (' ' != *b) *p++ = *b;

    /*** Print 2D data item and free temporary buffer. ***/

    printf("2D<%s>", buf + 1);
    free(buf);

    if (amatch && strlen(amatch)) {
/*
     printf("MATCH<%s>", amatch);
*/
     printf("ALAB<");
     for (p = amatch; p < amatch+strlen(amatch)-1; p++) {
      /*printf(".%.1s,",p);*/
      printf("%.1s,",p);
     }
     printf("%.1s>",p);
     /*printf(".%.1s>",p);*/
     free(amatch);
     amatch = NULL;
    }

    printf("|\n", buf + 1);

    /*** Deallocate mol (this deallocates all components, e.g. dep). ***/
    dt_dealloc(mol);
  }

  return 0;
}
